/* $Workfile: Dsp54Api.c $   $Revision: 11 $    $Modtime: 2/16/05 12:07p $  */
/*--------------------------------------------------------------------------*\

  Odin TeleSystems Inc.

  SAA-1007-1 OTX DSP C54x Software Development Kit (SDK)

  Implementation of the Standard API functions that should not be
  modified by the user.

\*--------------------------------------------------------------------------*/

#include <string.h>  /* for memset() */

#if BIOS_APPLICATION && CHIP_5416
#include "DspDemo1cfg.h"
#endif

#include <Dsp54Ioctl.h>

#include "OtxRev.h"

#include "Dsp54Api.h"

#ifdef MCBSP_SUPPORT
#include <csl.h>
#include <csl_mcbsp.h>
#include <csl_dma.h>
#include <csl_irq.h>
#include <csl_timer.h>
#endif

static void InitBufferedSerialPorts(void);


/*--------------------------------------------------------------------------*\
  Macros
\*--------------------------------------------------------------------------*/
/* @type C54X_SET_INT_MASK_BIT | Macro for Masking Interrupts */
#define C54X_SET_INT_MASK_BIT   asm("\t ssbx    INTM")
/* @type C54X_CLR_INT_MASK_BIT | Macro for Unmasking Interrupts */
#define C54X_CLR_INT_MASK_BIT   asm("\t rsbx    INTM")


/*--------------------------------------------------------------------------*\
  Global Data
\*--------------------------------------------------------------------------*/

// OTXDSP_HWY_SPEEDS could be specified on the command line, but if it is not we default to 2M support only
#ifndef OTXDSP_HWY_SPEEDS
unsigned short g_nSupportedHighwaySpeeds = OTXDSP_HWY_SPEED_2M; // Bit pattern, 0x0001=2M support only
#else
unsigned short g_nSupportedHighwaySpeeds = OTXDSP_HWY_SPEEDS;   // Bit pattern of supported highway speeds
#endif


volatile unsigned short *g_h2dCmdAddrP      = (unsigned short *) OTXDSP_SDK_IOCTL_CODE_ADDR;
volatile unsigned short *g_h2dHandleAddrP   = (unsigned short *) OTXDSP_SDK_IOCTL_HANDLE_ADDR;
volatile unsigned short *g_h2dTaskRefAddrP  = (unsigned short *) OTXDSP_SDK_IOCTL_TASKREF_ADDR;
volatile unsigned short *g_h2dDataLenAddrP  = (unsigned short *) OTXDSP_SDK_IOCTL_DATA_LEN_ADDR;
volatile unsigned short *g_h2dDataAddrP     = (unsigned short *) OTXDSP_SDK_IOCTL_DATA_ADDR;

static unsigned short *g_pCurrDspWriteAddr  = (unsigned short *) OTXDSP_SDK_EVENT_BUF_ADDR;

OtxDspHpiEventBufDiag g_hpiEventBufDiag;

OtxDspTimeStamp g_timeStamp;

// Set in OtxDspStartProgram()
unsigned short g_nHighwaySpeed;
unsigned short g_nTimeslotsPerFrame;
unsigned short g_nWordslotsPerFrame;

volatile short g_bGotRxSerialPortInterrupt = OTXDSP_FALSE;
volatile short g_bGotTxSerialPortInterrupt = OTXDSP_FALSE;
volatile short g_bGotHostPortInterrupt     = OTXDSP_FALSE;

volatile short g_bRxSerialPortHalf = OTXDSP_FALSE;
volatile short g_bTxSerialPortHalf = OTXDSP_FALSE;

volatile unsigned short g_nFrameBufferCounter = 0;
volatile unsigned short g_nTimerCounter = 0;

OtxDspIsrCallbackFnP g_pFnTimerIsr = NULL;
OtxDspIsrCallbackFnP g_pFnRxSerialPortIsr = NULL;
OtxDspIsrCallbackFnP g_pFnTxSerialPortIsr = NULL;
OtxDspIsrCallbackFnP g_pFnHostPortIsr = NULL;


/*--------------------------------------------------------------------------*\
  Certain Memory Mapped useful Registers
\*--------------------------------------------------------------------------*/
volatile unsigned short * const C54X_IMR_REG    = (volatile unsigned short *) 0x0000;
volatile unsigned short * const C54X_IFR_REG    = (volatile unsigned short *) 0x0001;
volatile unsigned short * const C54X_ST0_REG    = (volatile unsigned short *) 0x0006;
volatile unsigned short * const C54X_ST1_REG    = (volatile unsigned short *) 0x0007;
volatile unsigned short * const C54X_PMST_REG   = (volatile unsigned short *) 0x001D;
volatile unsigned short * const C54X_BDRR_REG   = (volatile unsigned short *) 0x0020;
volatile unsigned short * const C54X_BDXR_REG   = (volatile unsigned short *) 0x0021;
volatile unsigned short * const C54X_SPC0_REG   = (volatile unsigned short *) 0x0022;
volatile unsigned short * const C54X_TIM_REG    = (volatile unsigned short *) 0x0024;
volatile unsigned short * const C54X_PRD_REG    = (volatile unsigned short *) 0x0025;
volatile unsigned short * const C54X_TCR_REG    = (volatile unsigned short *) 0x0026;
volatile unsigned short * const C54X_TRCV_REG   = (volatile unsigned short *) 0x0030;
volatile unsigned short * const C54X_TDXR_REG   = (volatile unsigned short *) 0x0031;
volatile unsigned short * const C54X_SPC1_REG   = (volatile unsigned short *) 0x0032;
volatile unsigned short * const C54X_SWWSR_REG  = (volatile unsigned short *) 0x0028;
volatile unsigned short * const C54X_BSCR_REG   = (volatile unsigned short *) 0x0029;
volatile unsigned short * const C54X_HPIC_REG   = (volatile unsigned short *) 0x002C;
volatile unsigned short * const C54X_BSPCE_REG  = (volatile unsigned short *) 0x0023;
volatile unsigned short * const C54X_AXR_REG    = (volatile unsigned short *) 0x0038;
volatile unsigned short * const C54X_BKX_REG    = (volatile unsigned short *) 0x0039;
volatile unsigned short * const C54X_ARR_REG    = (volatile unsigned short *) 0x003A;
volatile unsigned short * const C54X_BKR_REG    = (volatile unsigned short *) 0x003B;
volatile unsigned short * const C54X_CLKMD_REG  = (volatile unsigned short *) 0x0058;


  

#if 1 // temp
unsigned short g_debugData[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
unsigned short g_debugBuf[32] = { 0, 0, 0, 0, 0, 0, 0, 0,
                                  0, 0, 0, 0, 0, 0, 0, 0,    
                                  0, 0, 0, 0, 0, 0, 0, 0,
                                  0, 0, 0, 0, 0, 0, 0, 0 };
#endif


#ifdef MCBSP_SUPPORT
TIMER_Config timerCfg = {
  0x0003, // tcr 
          /*
          TIMER_TCR_RMK(
          TIMER_TCR_SOFT_WAITZERO,
          TIMER_TCR_FREE_NOSOFT,
          TIMER_TCR_TRB_RESET,
          TIMER_TCR_TSS_START,
          TIMER_TCR_TDDR_OF(0)
          ),                        // TCR0 
          */
          
          0xFFFF /* prd */
};

TIMER_Handle hTimer;

DMA_Handle hDma0Rx = 0;
DMA_Handle hDma1Tx = 0;
MCBSP_Handle hMcbsp0 = 0;
interrupt void dmaCha0Isr(void);
interrupt void dmaCha1Isr(void);

MCBSP_Config mcbspCfg0= {
  MCBSP_SPCR1_RMK(
    MCBSP_SPCR1_DLB_OFF,                   // DLB    = 0 
    MCBSP_SPCR1_RJUST_RZF,                 // RJUST  = 0 
    MCBSP_SPCR1_CLKSTP_DISABLE,            // CLKSTP = 0 
    MCBSP_SPCR1_DXENA_NA,                  // DXENA  = 0 
#if (_MCBSP_ABIS_SUPPORT)
    MCBSP_SPCR1_ABIS_DISABLE,              // ABIS   = 0 
#endif
    MCBSP_SPCR1_RINTM_RRDY,                // RINTM  = 0 
    MCBSP_SPCR1_RRST_DISABLE               // RRST   = 0 
  ),
  MCBSP_SPCR2_RMK(
    MCBSP_SPCR2_FREE_NO,                   // FREE   = 0 
    MCBSP_SPCR2_SOFT_NO,                   // SOFT   = 0 
    0, //MCBSP_SPCR2_FRST_DEFAULT,                  // FRST   = 0
    0, //MCBSP_SPCR2_GRST_DEFAULT,                 // GRST   = 0 
    MCBSP_SPCR2_XINTM_XRDY,                // XINTM  = 0 
    MCBSP_SPCR2_XRST_DISABLE               // XRST   = 0
  ),
  MCBSP_RCR1_RMK( 
    MCBSP_RCR1_RFRLEN1_OF(15),              // 16 words per frame 
    MCBSP_RCR1_RWDLEN1_16BIT                // 16 bits per word (2 timeslots)
  ),
  MCBSP_RCR2_RMK(    
    MCBSP_RCR2_RPHASE_SINGLE,              // RPHASE  = 0; Single phase 
    MCBSP_RCR2_RFRLEN2_OF(0),              // RFRLEN2 = 0; Not used 
    MCBSP_RCR2_RWDLEN2_DEFAULT,            // RWDLEN2 = 0; Not used 
    MCBSP_RCR2_RCOMPAND_MSB,               // RCOMPAND = 0; No companding, MSB first
    MCBSP_RCR2_RFIG_NO,                   // RFIG    = 0, Restart frame on any sync pulse (set RSYNCERR if error) 
    MCBSP_RCR2_RDATDLY_0BIT                // RDATDLY = 0 
  ),  
  MCBSP_XCR1_RMK(    
    MCBSP_XCR1_XFRLEN1_OF(15),             // 16 words per frame   
    MCBSP_XCR1_XWDLEN1_16BIT               // 16 bits per word (2 timeslots) 
  ),   
  MCBSP_XCR2_RMK(   
    MCBSP_XCR2_XPHASE_SINGLE,              // XPHASE  = 0; Single phase 
    MCBSP_XCR2_XFRLEN2_OF(0),              // XFRLEN2 = 0; Not used 
    MCBSP_XCR2_XWDLEN2_DEFAULT,            // XWDLEN2 = 0; Not used 
    MCBSP_XCR2_XCOMPAND_MSB,               // XCOMPAND = 0; No companding, MSB first
    MCBSP_XCR2_XFIG_NO,                   // XFIG    = 0, Restart frame on any sync pulse 
    MCBSP_XCR2_XDATDLY_1BIT                // XDATDLY = 1; 1 bit delay after the frame sync 
  ),            
  MCBSP_SRGR1_DEFAULT,   
  MCBSP_SRGR2_DEFAULT,  
  MCBSP_MCR1_DEFAULT,
  MCBSP_MCR2_DEFAULT, 
  MCBSP_PCR_RMK(
    MCBSP_PCR_XIOEN_SP,                     // XIOEN    = 0   
    MCBSP_PCR_RIOEN_SP,                     // RIOEN    = 0   
    MCBSP_PCR_FSXM_EXTERNAL,                // FSXM     = 0   
    MCBSP_PCR_FSRM_EXTERNAL,                // FSRM     = 0   
#if (_MCBSP_SCLKME_SUPPORT)
    MCBSP_PCR_SCLKME_NO,                    // SCLKME   = 0   
#endif
    MCBSP_PCR_CLKXM_INPUT,                  // CLKXM    = 0   
    MCBSP_PCR_CLKRM_INPUT,                  // CLKRM    = 0   
    MCBSP_PCR_FSXP_ACTIVEHIGH,              // FSXP     = 0   
    MCBSP_PCR_FSRP_ACTIVEHIGH,              // FSRP     = 0   
    MCBSP_PCR_CLKXP_RISING,                 // CLKXP      
    MCBSP_PCR_CLKRP_RISING                  // CLKRP       
  ), 
#if (!_MCBSP_128_CH_ENABLE_SUPPORT)
  0x0000u,                          /* RCERA */
  0x0000u,                          /* RCERB */
  0x0000u,                          /* XCERA */
  0x0000u                           /* XCERB */
#else
  MCBSP_RCERA_DEFAULT, 
  MCBSP_RCERB_DEFAULT, 
  MCBSP_RCERC_DEFAULT, 
  MCBSP_RCERD_DEFAULT, 
  MCBSP_RCERE_DEFAULT, 
  MCBSP_RCERF_DEFAULT, 
  MCBSP_RCERG_DEFAULT, 
  MCBSP_RCERH_DEFAULT, 
  MCBSP_XCERA_DEFAULT,
  MCBSP_XCERB_DEFAULT,
  MCBSP_XCERC_DEFAULT,
  MCBSP_XCERD_DEFAULT,  
  MCBSP_XCERE_DEFAULT,
  MCBSP_XCERF_DEFAULT,  
  MCBSP_XCERG_DEFAULT,
  MCBSP_XCERH_DEFAULT
#endif
};     

DMA_Config  dmaCfg0Rx = {
  0,                                   // Priority 
#if (_DMA_CH_RLDR_SUPPORT)
  0,                                   // Autoix 
#endif
  DMA_DMMCR_RMK(
    DMA_DMMCR_AUTOINIT_ON, // OFF
    DMA_DMMCR_DINM_ON,
    DMA_DMMCR_IMOD_HALF_AND_FULL,
    DMA_DMMCR_CTMOD_ABU,
    DMA_DMMCR_SLAXS_OFF,
    DMA_DMMCR_SIND_NOMOD,
    DMA_DMMCR_DMS_DATA,
    DMA_DMMCR_DLAXS_OFF,
    DMA_DMMCR_DIND_DMIDX0,
    DMA_DMMCR_DMD_DATA
  ),                                   // DMMCR 
  DMA_DMSFC_RMK(
    DMA_DMSFC_DSYN_REVT0,
    DMA_DMSFC_DBLW_OFF,
    DMA_DMSFC_FRAMECNT_OF(0)           // The frame count register does not have a function in ABU mode
  ),                                   // DMSFC
  (DMA_AdrPtr)MCBSP_ADDR(DRR10),       // DMSRC
  (DMA_AdrPtr)(OTXDSP_BSP_RXBUF_ADDR), // DMDST
  (Uint16)(OTXDSP_BSP_RXBUF_SIZE)    // DMCTR = buffsize 
};


DMA_Config  dmaCfg1Tx = {
  0,                                   // Priority
#if (_DMA_CH_RLDR_SUPPORT)
  0,                                   // Autoix   
#endif
  DMA_DMMCR_RMK(
    DMA_DMMCR_AUTOINIT_ON,  // OFF
    DMA_DMMCR_DINM_ON,   
    DMA_DMMCR_IMOD_HALF_AND_FULL,
    DMA_DMMCR_CTMOD_ABU,
    DMA_DMMCR_SLAXS_OFF,
    DMA_DMMCR_SIND_DMIDX0,
    DMA_DMMCR_DMS_DATA,
    DMA_DMMCR_DLAXS_OFF,
    DMA_DMMCR_DIND_NOMOD,
    DMA_DMMCR_DMD_DATA
  ),                                   // DMMCR 
  DMA_DMSFC_RMK(
    DMA_DMSFC_DSYN_XEVT0,
    DMA_DMSFC_DBLW_OFF,
    DMA_DMSFC_FRAMECNT_OF(0)           // The frame count register does not have a function in ABU mode
  ),                                   // DMSFC
  (DMA_AdrPtr)(OTXDSP_BSP_TXBUF_ADDR), // DMSRC 
  (DMA_AdrPtr)MCBSP_ADDR(DXR10),       // DMDST 
  (Uint16)(OTXDSP_BSP_TXBUF_SIZE)    // DMCTR = buffsize 
};
#endif // MCBSP_SUPPORT



/*--------------------------------------------------------------------------*\
  Interrupt Service Routines
\*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*\
  @func
    TIMER INTERRUPT SERVICE ROUTINE

    This interrupt service routine will be executed every time TIM 
    decrements to 0.

    The timer interrupt is used to create an accurate timestamp 
    (64 bits) for notify events to be sent to the host. The timer interrupt
    frequency is 10.000MHz for ASM and PCI boards and 10.036MHz for PCMCIA boards.

\*--------------------------------------------------------------------------*/
interrupt void c_int19(
  void
)
{

  if (g_timeStamp.nLowPart == 0xFFFF) {
    if (g_timeStamp.nMidPart == 0xFFFF) {
      g_timeStamp.nHighPart++;
    }
    g_timeStamp.nMidPart++;
  }
  g_timeStamp.nLowPart++;

  if (g_pFnTimerIsr) {
    (*g_pFnTimerIsr)();
  }

}



#ifdef MCBSP_SUPPORT
/*--------------------------------------------------------------------------*\
  @func
    Interrupt service routine for the REVT DMA interrupts
\*--------------------------------------------------------------------------*/
interrupt void dmaCha0Isr(void) 
{
  g_debugData[2]++;

  g_nFrameBufferCounter++;

#ifdef OTXDSP_SDK_EVENT_BUF_INCLUDE_TIMESTAMP
  g_nTimerCounter = 0; // Functionality not supported when the timer is used for timestamps
#else
  TIMER_stop(hTimer);  
  g_nTimerCounter = TIMER_RGET(TIM0);
  TIMER_reload(hTimer);  
  TIMER_start(hTimer);  
#endif

  //g_debugData[5] = DMA_RGET(DMIDX0);

  g_bRxSerialPortHalf = !g_bRxSerialPortHalf;

  g_bGotRxSerialPortInterrupt = OTXDSP_TRUE;
#if BIOS_APPLICATION
  SEM_ipost(&semProcessing);
#else
  if (g_pFnRxSerialPortIsr) {
    (*g_pFnRxSerialPortIsr)();
  }
#endif
    
}
#endif // MCBSP_SUPPORT


#ifdef MCBSP_SUPPORT
/*--------------------------------------------------------------------------*\
  @func
    Interrupt service routine for the XEVT DMA interrupts
\*--------------------------------------------------------------------------*/
interrupt void dmaCha1Isr(void) {
  g_debugData[3]++;

  g_bTxSerialPortHalf = !g_bTxSerialPortHalf;

  g_bGotTxSerialPortInterrupt = OTXDSP_TRUE;
#if BIOS_APPLICATION
  SEM_ipost(&semProcessing);
#else
  if (g_pFnTxSerialPortIsr) {
    (*g_pFnTxSerialPortIsr)();
  }
#endif

}
#endif // MCBSP_SUPPORT

/*--------------------------------------------------------------------------*\
  @func
    Serial Port Receive Interrupt Service Routine

    The DSP is connected to the system highway (32 timeslots).

    We will get this interrupt every 25 highway frames (since the
    autobuffer is 800 words long, which is 50 frames (50*16=800), and we
    get this interrupt when the autobuffer is half full and also when is
    completely full).

    This function is not used is the DSP is equipped wich McBSP serial ports.

\*--------------------------------------------------------------------------*/
interrupt void c_int20(
  void
)
{
#ifndef MCBSP_SUPPORT
#if 1
  g_nFrameBufferCounter++;

#ifdef OTXDSP_SDK_EVENT_BUF_INCLUDE_TIMESTAMP
  g_nTimerCounter = 0; // Functionality not supported when the timer is used for timestamps
#else
  *C54X_TCR_REG = 0x0017;    // Turn timer off 
  g_nTimerCounter = *C54X_TIM_REG;
  *C54X_TCR_REG = 0x0037;    // Reload timer
  *C54X_TCR_REG = 0x0007;    // Turn timer on 
#endif

  *C54X_BSPCE_REG |= 0x2000;   // Set BRE=1 to enable autobuffering again (since we are running with HALTR=1)

  if (*C54X_BSPCE_REG & 0x4000) {
    // BSPCE RH=1 --> We should read the second half
    g_bRxSerialPortHalf = OTXDSP_TRUE;
  } else {
    // BSPCE RH=0 --> We should read the first half
    g_bRxSerialPortHalf = OTXDSP_FALSE;
  }

  g_bGotRxSerialPortInterrupt = OTXDSP_TRUE;

  if (g_pFnRxSerialPortIsr) {
    (*g_pFnRxSerialPortIsr)();
  }

#else // Just for testing
  Uint16 readVal;
  //g_debugData[0]++;

  readVal = MCBSP_read16(hMcbsp0);
#endif
#endif // !MCBSP_SUPPORT
}


/*--------------------------------------------------------------------------*\
  @func
    Serial Port Transmit Interrupt Service Routine

    This function is not used is the DSP is equipped wich McBSP serial ports.

\*--------------------------------------------------------------------------*/
interrupt void c_int21(
  void
)
{
#ifndef MCBSP_SUPPORT
#if 1
  *C54X_BSPCE_REG |= 0x0400;   // Set BXE=1 to enable autobuffering again (since we are running with HALTX=1)

  if (*C54X_BSPCE_REG & 0x0800) {
    // BSPCE XH=1 --> We should fill the second half
    g_bTxSerialPortHalf = OTXDSP_TRUE;
  } else {
    // BSPCE XH=0 --> We should fill the first half
    g_bTxSerialPortHalf = OTXDSP_FALSE;
  }

  g_bGotTxSerialPortInterrupt = OTXDSP_TRUE;

  if (g_pFnTxSerialPortIsr) {
    (*g_pFnTxSerialPortIsr)();
  }


#else // Just for testing
  //g_debugData[1]++;

  MCBSP_write16(hMcbsp0, 5);
#endif
#endif // !MCBSP_SUPPORT
}




/*--------------------------------------------------------------------------*\
  @func
\*--------------------------------------------------------------------------*/
interrupt void c_int22(
  void
)
{
  // Nothing
}


/*--------------------------------------------------------------------------*\
  @func
\*--------------------------------------------------------------------------*/
interrupt void c_int23(
  void
)
{
 // Nothing
}



/*--------------------------------------------------------------------------*\
  @func
    Host Port Interrupt. Activated when the host wants our attention.
\*--------------------------------------------------------------------------*/
interrupt void c_int25(
  void
)
{
  g_bGotHostPortInterrupt = OTXDSP_TRUE;
#if BIOS_APPLICATION
  SEM_ipost(&semProcessing);
#else
  if (g_pFnHostPortIsr) {
    (*g_pFnHostPortIsr)();
  }
#endif

}



/*--------------------------------------------------------------------------*\
  API Functions
\*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*\
  @func
    Run Standard OTX DSP SDK API Initializations
\*--------------------------------------------------------------------------*/
void OtxDspInit(
  void
)
{

  // Configure the Processor
  OtxDspInitProcessor();

  // Initialize Standard Global Data
  OtxDspInitGlobalData();

  // Initialize the Host Port Interface
  OtxDspInitHostPortInterface();

  // Initialize Interrupts. Mask unwanted interrupts
  OtxDspInitInterrupts(0x0200);  // 0x0200=allow Host Port Interrupt

  C54X_CLR_XF_BIT;   // Turn LED on (just as an indication of executing this piece of code)

}

/*--------------------------------------------------------------------------*\
  @func
    Sets up the processor to the desired operating mode.
\*--------------------------------------------------------------------------*/
void OtxDspInitProcessor(
  void
)
{
  //C54X_CLR_XF_BIT;   // Turn LED on (just as an indication of executing this piece of code)

#if CHIP_5416
  *C54X_PMST_REG |= 0x0008;  // Set DROM=1 
#endif

#if CHIP_5410A
  *C54X_PMST_REG |= 0x0008;  // Set DROM=1 
#endif


#ifdef MCBSP_SUPPORT
  CSL_init();
#endif // MCBSP_SUPPORT

#if !CHIP_54CST
  *C54X_IMR_REG = 0;      // Disable all interrupts for now
#endif


#if CHIP_54CST
  // User code can start at 0xC80 on the 54CST (if the CST ROM code is to be used)
  // Therefore we put the interrupt vector at the beginning of the user code area
  memset((unsigned short *) 0x0C80, 0, 128);
  IRQ_setVecs(0x0C80);
#elif CHIP_5410A
  memset((unsigned short *) 0x0180, 0, 128);
  IRQ_setVecs(0x0180);
#elif CHIP_5416
  memset((unsigned *) 0x0180, 0, 128);
  IRQ_setVecs(0x0180);
#else
  *C54X_PMST_REG &= ~0xFF80; // Clear Interrupt vector pointer
  *C54X_PMST_REG |=  0x0180; // Set Interrupt vector pointer to 0x180 (as specified in the *.CMD file)
#endif

#if CHIP_5410A || CHIP_54CST || CHIP_5416
  *C54X_SWWSR_REG = 0        // Set Software Wait States
  | (1 << 15) // Extended program address control bit. Extends the address range for the program space wait states.
  | (0 << 12) // Wait states for external I/O space 0x0000-0xFFFF (not connected)
  | (1 << 9)  // Wait states for external data space 0x8000-0xFFFF (internal memory if DROM=1)
  | (1 << 6)  // Wait states for external data space 0x0000-0x7FFF (internal memory)
  | (1 << 3)  // Wait states for external program space 0x8000-0xFFFF 
  | (1 << 0)  // Wait states for external program space 0x0000-0x7FFF 
  ;
#else // 549/548
  *C54X_SWWSR_REG = 0        // Set Software Wait States
  | (0 << 12) // Wait states for I/O space 0x0000-0xFFFF (not connected)
  | (2 << 9)  // Wait states for data space 0x8000-0xFFFF (external memory, if available)
  | (0 << 6)  // Wait states for data space 0x0000-0x7FFF (internal memory)
  | (2 << 3)  // Wait states for program space 0x8000-0xFFFF (external memory, if available)
  | (0 << 0)  // Wait states for program space 0x0000-0x7FFF (internal memory)
  ;
#endif

  //*C54X_BSCR_REG = 0x0;        

  *C54X_CLKMD_REG = 0x5320;  // Set CLK multiplier to 6 (CLKINmax = 13.333MHz), with a 160 lockup time
  *C54X_CLKMD_REG = 0x5322;  // Start lockup sequence
  while (!(*C54X_CLKMD_REG & 0x0001)); // Wait for lockup sequence to complete

#ifdef MCBSP_SUPPORT  
  IRQ_plug(IRQ_EVT_TINT0, &c_int19);
#endif  


}


/*--------------------------------------------------------------------------*\
  @func
    Initalizes the global data used by the Standard API functions.
\*--------------------------------------------------------------------------*/
void OtxDspInitGlobalData(
  void
)
{
  unsigned short *pTemp;
  short i;

  g_pCurrDspWriteAddr  = (unsigned short *) OTXDSP_SDK_EVENT_BUF_ADDR;

  g_hpiEventBufDiag.cwFreeBuffer = g_hpiEventBufDiag.cwMinFreeBuffer = OTXDSP_SDK_EVENT_BUF_LEN;
  g_hpiEventBufDiag.coLostEvents = 0;  

  memset(&g_timeStamp, 0, sizeof(OtxDspTimeStamp));

  pTemp = (unsigned short *) OTXDSP_BSP_RXBUF_ADDR; // Autobuffer receive area
  for (i = 0; i < OTXDSP_BSP_RXBUF_SIZE; i++) {
    *pTemp = 0;
    pTemp++;
  }

  pTemp = (unsigned short *) OTXDSP_BSP_TXBUF_ADDR; // Autobuffer transmit area
  for (i = 0; i < OTXDSP_BSP_TXBUF_SIZE; i++) {
    *pTemp = 0;
    pTemp++;
  }

}


/*--------------------------------------------------------------------------*\
  @func
    Initialize and unmask specified interrupts.
\*--------------------------------------------------------------------------*/
void OtxDspInitInterrupts(
  unsigned short valueIMR       /* @parm Interrupts to be unmasked */
)
{
  C54X_SET_INT_MASK_BIT;   // Globally disable interrupts by setting the INTM bit 

  *C54X_IFR_REG = valueIMR;          /* Clear any pending interrupts for this interrupt(s) */

  *C54X_IMR_REG |= valueIMR;         /* Allow this interrupt(s) */

  C54X_CLR_INT_MASK_BIT;  // Globally enable interrupts by clearing the INTM bit
}


/*--------------------------------------------------------------------------*\
  @func
    Initialize the Host Port Interface
\*--------------------------------------------------------------------------*/
void OtxDspInitHostPortInterface(
  void
)
{
  asm("\t STM    #02h, hpic");  /* Switch to SAM mode (if we are not already in that mode) */
  C54X_NOP;                          /* This can't be IDLE2/3 instruction. See Chapter 8.5.5 in the CPU and Peripherals Reference Book) */
  C54X_NOP;                          /* A few extra can't hurt */
  C54X_NOP;         
                   /* A few extra can't hurt */
#ifdef MCBSP_SUPPORT  
  IRQ_plug(IRQ_EVT_HPINT, &c_int25);
#endif  
}





/*--------------------------------------------------------------------------*\
  @func
    Reset the Buffered Serial Ports (BSP)
\*--------------------------------------------------------------------------*/
void OtxDspResetBufferedSerialPorts(
  void
)
{
#ifdef MCBSP_SUPPORT
  //MCBSP_reset(hMcbsp0);
#else
  *C54X_SPC0_REG = 0x0000;      /* Reset BSP */
#endif
}


/*--------------------------------------------------------------------------*\
  @func
    Initialize the Buffered Serial Ports (BSP)
\*--------------------------------------------------------------------------*/
void OtxDspInitBufferedSerialPorts(
  void
)
{
#ifdef MCBSP_SUPPORT
  // Nothing 
#else
  InitBufferedSerialPorts();
#endif
}

/*--------------------------------------------------------------------------*\
  @func
    Initialize the Buffered Serial Ports (BSP)
\*--------------------------------------------------------------------------*/
void InitBufferedSerialPorts(
  void
)
{
#ifdef MCBSP_SUPPORT
  Uint16 cha0EventId, cha1EventId;
  //static short g_bMcBspInitalized = OTXDSP_FALSE;

  //if (g_bMcBspInitalized) {
  //  return;
  //}

  /* Temporarily disable all maskable interrupts */ 
  IRQ_globalDisable();

  g_nFrameBufferCounter = 0;

  /* Open MCBSP channel 0 */
  hMcbsp0 = MCBSP_open(MCBSP_PORT0, MCBSP_OPEN_RESET); 
  
  /* Write configuration values to MCBSP 0 control registers */
  MCBSP_config(hMcbsp0, &mcbspCfg0);       

  /* Open DMA channels 0 & 1 */
  hDma0Rx = DMA_open(DMA_CHA0, 0); 
  hDma1Tx = DMA_open(DMA_CHA1, 0);

  /* Write configuration structure values to DMa control registers */
  DMA_config(hDma0Rx, &dmaCfg0Rx);      
  DMA_config(hDma1Tx, &dmaCfg1Tx);     

  /* Get Event ID's associated with DMA channels */
  cha0EventId = DMA_getEventId(hDma0Rx);
  cha1EventId = DMA_getEventId(hDma1Tx); 
 
  /* Enable mask of DMA channel 0 and 1 interrupts by setting */
  /* the interrupt selector value in DMA priority and enable  */
  /* register, DMPREC                                         */
  DMA_FSET(DMPREC,INTOSEL,DMA_DMPREC_INTOSEL_CH0_TO_CH5);

  /* Set FREE bit in DMPREC to enable free run on DMA when emulator is suspended        */
  //DMA_FSET(DMPREC,FREE,1);

  //DMA_FSET(DMPREC,AUTOIX, DMA_DMPREC_AUTOIX_USE_CHAN);


  /* Clear all pending maskable interrupts */
#if 1
  CHIP_RSET(IFR,0xFFFFu);
#else
  //IRQ_clear(cha0EventId);
  //IRQ_clear(cha1EventId);
#endif


  /* Enable interrupts for DMA channels 0 & 1 */
  IRQ_enable(cha0EventId);
  IRQ_enable(cha1EventId);

  /* Place ISR addresses at associated vector location */
  IRQ_plug(cha0EventId, &dmaCha0Isr);
  IRQ_plug(cha1EventId, &dmaCha1Isr);

  /* Set value of Global Index Register , DMIDX0 */
  DMA_RSET(DMIDX0,1);  // Or 0 ??

  g_bRxSerialPortHalf = OTXDSP_TRUE;
  g_bTxSerialPortHalf = OTXDSP_TRUE;
           
  /* Start DMA */
  DMA_start(hDma1Tx);
  DMA_start(hDma0Rx); 

  //IRQ_enable(IRQ_EVT_RINT0);
  //IRQ_enable(IRQ_EVT_XINT0);

  //IRQ_plug(IRQ_EVT_RINT0, &c_rint0);
  //IRQ_plug(IRQ_EVT_XINT0, &c_xint0);

  /* Enable all maskable interrupts, INTM = 0 */
  IRQ_globalEnable();

  /* Take MCBSP transmit and receive out of reset */  
  MCBSP_start(hMcbsp0, 
              MCBSP_RCV_START|MCBSP_XMIT_START,
              0
              );

  //g_debugData[5] = MCBSP_read16(hMcbsp0);

  /* Prime MCBSP DXR */
  //while(!(MCBSP_xrdy(hMcbsp0))){
  //  ;
  //}
  MCBSP_write16(hMcbsp0,0);
 
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */

  /* 
  g_debugBuf[0] = MCBSP_RGET(SPCR10);
  g_debugBuf[1] = MCBSP_RGET(SPCR20);
  g_debugBuf[2] = MCBSP_RGET(RCR10);
  g_debugBuf[3] = MCBSP_RGET(RCR20);
  g_debugBuf[4] = MCBSP_RGET(XCR10);
  g_debugBuf[5] = MCBSP_RGET(XCR20);
  g_debugBuf[6] = MCBSP_RGET(SRGR10);
  g_debugBuf[7] = MCBSP_RGET(SRGR20);
  g_debugBuf[8] = MCBSP_RGET(MCR10);
  g_debugBuf[9] = MCBSP_RGET(MCR20);
  g_debugBuf[10] = MCBSP_RGET(PCR0);
  g_debugData[4] = *C54X_IMR_REG;   
  */

  //g_bMcBspInitalized = OTXDSP_TRUE;




#else  // !MCBSP_SUPPORT

  *C54X_SPC0_REG = 0x0000;      /* Reset BSP */

  *C54X_SPC0_REG = 0x0000;      /* reset SP0                            */
                                /*     =0 (b0) : reserved, always zero  */
                                /* DBL =0 (b1) : loopback mode disabled */
                                /* FO  =0 (b2) : 16-bit word operation  */
                                /* FSM =1 (b3) : frame sync control     */
                                /* MCM =0 (b4) : external CLKX          */
                                /* TXM =0 (b5) : external FSK           */
                                /* XRST=0 (b6) : Xmit reset             */
                                /* RRST=0 (b7) : Recv reset             */

  g_nFrameBufferCounter = 0;


  OtxDspInitInterrupts(0x0030);  /* Clear pending TX and RX interrupts and enable them */


  *C54X_BSPCE_REG = 0xB400;     /*     BSP Extended functions register */
                                /*  0  CLKDV=0  (b0-4) : internal transmit clock division factor */
                                /*  0  FSP  =0  (b5) : frame sync polarity */
                                /*  0  CLKP =0  (b6) : clock polarity */
                                /*  0  FE   =0  (b7) : format extension */
                                /*  1  FIG  =0  (b8) : frame ignore */
                                /*  0  PCM  =0  (b9) : PCM mode */
                                /*  0  BSP Autobuffering Unit bits */
                                /*  1  BXE  =1 (b10) : autobuffering transmit enable */
                                /*  0  XH   =0 (b11) : transmit buffer half transmitted */
                                /*  1  HALTX=0 (b12) : autobuffering transmit halt */
                                /*  1  BRE  =1 (b13) : autobuffering receive enable */
                                /*  0  RH   =0 (b14) : receive buffer half transmitted */
                                /*  1  HALTR=0 (b15) : autobuffering receive halt */

  *C54X_BKX_REG = OTXDSP_BSP_TXBUF_SIZE;          /* Size of transmit buffer */
  *C54X_AXR_REG = OTXDSP_BSP_TXBUF_ADDR - 0x800;  /* Relative start address of transmit buffer (0x0800 + AXR) */
  *C54X_BKR_REG = OTXDSP_BSP_RXBUF_SIZE;          /* Size of receive buffer */
  *C54X_ARR_REG = OTXDSP_BSP_RXBUF_ADDR - 0x800;  /* Relative start address of receive buffer (0x0800 + ARR) */

  *C54X_SPC0_REG = 0x00C0;      /* take SP0 out of reset                */
                                /* XRST=0 (b6) : Xmit reset             */
                                /* RRST=0 (b7) : Recv reset             */

#endif // MCBSP_SUPPORT
}



/*--------------------------------------------------------------------------*\
  @func
\*--------------------------------------------------------------------------*/
short OtxDspStartProgram(
  OtxDspSdkIoInStartProgramS *pSin
)
{

  // Check if we are in DIV mode or PLL mode
  if (*C54X_CLKMD_REG & 0x0001) {
    // Since we are already in PLL mode (we are just changing the multiplier), we
    // must first switch over to DIV mode before we switch back to the new PLL
    // mode.
    *C54X_CLKMD_REG = 0;  // switch to DIV mode
    while (*C54X_CLKMD_REG & 0x0001); // Wait until the DSP has switched over to DIV mode
    *C54X_CLKMD_REG = 0;  // Reset PLLON/OFF now when status is in DIV mode
  }

  *C54X_CLKMD_REG = pSin->nClkMdVal; // Reconfigure PLLMULL, PLLCOUNT, and PLLDIV
  *C54X_CLKMD_REG = pSin->nClkMdVal | 0x0002; // Set PLLNDIV high to start lockup sequence
  while (!(*C54X_CLKMD_REG & 0x0001)); // Wait for lockup sequence to complete


  // Check that the driver revision matches the revision of the DSP SDK
  if (pSin->nDrvRevNoMajor != OTX_DRV_REV_NO_MAJOR) {
    return OTXDSP_SDK_E_UNSUPPORTED_VERSION;
  }
  if (pSin->nDrvRevNoMinor != OTX_DRV_REV_NO_MINOR) {
    return OTXDSP_SDK_E_UNSUPPORTED_VERSION;
  }
  if (pSin->nDrvRevNoPr != OTX_DRV_REV_NO_PR) {
    return OTXDSP_SDK_E_UNSUPPORTED_VERSION;
  }
  if (pSin->nDrvRevNoPr == 0) {
    if (pSin->nDrvRevNoPversion != OTX_DRV_REV_NO_PVERSION) {
      return OTXDSP_SDK_E_UNSUPPORTED_VERSION;
    }
  }

  g_nHighwaySpeed      = pSin->nHighwaySpeed;
  g_nTimeslotsPerFrame = g_nHighwaySpeed >> 6;
  g_nWordslotsPerFrame = g_nHighwaySpeed >> 7;
#if 1 // temp
  g_debugData[4] = g_nHighwaySpeed;
  g_debugData[5] = g_nTimeslotsPerFrame;
  g_debugData[6] = g_nWordslotsPerFrame;
#endif

  if (pSin->nHighwaySpeed == 4096) {
    if (!(g_nSupportedHighwaySpeeds & OTXDSP_HWY_SPEED_4M)) {
      return OTXDSP_SDK_E_NOT_SUPPORTED;
    }
#ifdef MCBSP_SUPPORT
    mcbspCfg0.rcr1 = 
      MCBSP_RCR1_RMK( 
      MCBSP_RCR1_RFRLEN1_OF(31),              // 32 words (64 timeslots) per frame 
      MCBSP_RCR1_RWDLEN1_16BIT                // 16 bits per word (2 timeslots)
      );
    mcbspCfg0.xcr1 = 
      MCBSP_XCR1_RMK(    
      MCBSP_XCR1_XFRLEN1_OF(31),             // 32 words per frame   
      MCBSP_XCR1_XWDLEN1_16BIT               // 16 bits per word (2 timeslots) 
      );   
    if (g_nSupportedHighwaySpeeds & OTXDSP_HWY_SPEED_8M) {
      dmaCfg0Rx.dmctr = (Uint16)(OTXDSP_BSP_RXBUF_SIZE) >> 1;  // Divide by 2;
      dmaCfg1Tx.dmctr = (Uint16)(OTXDSP_BSP_TXBUF_SIZE) >> 1;  // Divide by 2;
    }

#endif
  } else if (pSin->nHighwaySpeed == 8192) {
    if (!(g_nSupportedHighwaySpeeds & OTXDSP_HWY_SPEED_8M)) {
      return OTXDSP_SDK_E_NOT_SUPPORTED;
    }
#ifdef MCBSP_SUPPORT
    mcbspCfg0.rcr1 = 
      MCBSP_RCR1_RMK( 
      MCBSP_RCR1_RFRLEN1_OF(63),              // 64 words (128 timeslots) per frame 
      MCBSP_RCR1_RWDLEN1_16BIT                // 16 bits per word (2 timeslots)
      );
    mcbspCfg0.xcr1 = 
      MCBSP_XCR1_RMK(    
      MCBSP_XCR1_XFRLEN1_OF(63),             // 64 words (128 timeslots) per frame   
      MCBSP_XCR1_XWDLEN1_16BIT               // 16 bits per word (2 timeslots) 
      );   
#endif
  } else if (pSin->nHighwaySpeed == 2048) {
    // Set by default
#ifdef MCBSP_SUPPORT
    if (g_nSupportedHighwaySpeeds & OTXDSP_HWY_SPEED_8M) {
      dmaCfg0Rx.dmctr = (Uint16)(OTXDSP_BSP_RXBUF_SIZE) >> 2;  // Divide by 4;
      dmaCfg1Tx.dmctr = (Uint16)(OTXDSP_BSP_TXBUF_SIZE) >> 2;  // Divide by 4;
    } else if (g_nSupportedHighwaySpeeds & OTXDSP_HWY_SPEED_4M) {
      dmaCfg0Rx.dmctr = (Uint16)(OTXDSP_BSP_RXBUF_SIZE) >> 1;  // Divide by 2;
      dmaCfg1Tx.dmctr = (Uint16)(OTXDSP_BSP_TXBUF_SIZE) >> 1;  // Divide by 2;
    }
#endif
  } else {
    return OTXDSP_SDK_E_INVALIDARG;
  }

  // Initialize the Buffered Serial Ports
  InitBufferedSerialPorts();

#ifdef MCBSP_SUPPORT
  g_debugData[0] = dmaCfg0Rx.dmctr;
  g_debugData[1] = dmaCfg1Tx.dmctr;
#endif

  return OTXDSP_SDK_S_OK;

}

/*--------------------------------------------------------------------------*\
  @func
\*--------------------------------------------------------------------------*/
short OtxDspInitTimer(
  void
)
{
#ifdef MCBSP_SUPPORT
  int old_intm;
  Uint16 eventId;

  // Temporarily disable all maskable interrupts, preserving 
  // previous state of INTM                                  
  old_intm = IRQ_globalDisable();

  hTimer = TIMER_open(TIMER_DEV0, TIMER_OPEN_RESET);

#if 0 
  TIMER_config(hTimer, &timerCfg);
#else // Alternative way. Configure PRD and TCR registers manually
  TIMER_RSET(PRD0, 0xFFFF);
  TIMER_RSET(TCR0, 0x0007); // turn timer on; Divide-down-ratio=8 => For a 160MHz CLKOUT [T=6.25ns] => 8*6.25ns = 50ns "ticks"
#endif

  /* Get Event ID associated with Timer interrupt */ 
  eventId = TIMER_getEventId(hTimer);    

  /* Clear any pending Timer interrupts */
  IRQ_clear(eventId);                     

  /* Place event ID in interrupt map for dispatcher */
  IRQ_map(eventId);

  /* Enable Timer interrupt */
  IRQ_enable(eventId);


  TIMER_start(hTimer);  

  // Restore old value of INTM
  IRQ_globalRestore(old_intm);

#else
  C54X_SET_INT_MASK_BIT;   // Globally disable interrupts by setting the INTM bit 

  *C54X_TCR_REG = 0x0010;    // Turn timer off 

  *C54X_TIM_REG = 0xFFFF;    // Timer counts down so put maximum value in the timer

  memset(&g_timeStamp, 0, sizeof(OtxDspTimeStamp));

  OtxDspInitInterrupts(0x0008);     // Clear pending TINT interrupts and enable them (and enable global interrupts)

  *C54X_TCR_REG = 0x0007;    // turn timer on; Divide-down-ratio=8 => For a 80MHz CLKOUT [T=12.5ns] => 8*12.5ns = 100ns "ticks"
#endif

  return OTXDSP_SDK_S_OK;
}



/*--------------------------------------------------------------------------*\
  @func
    Completes an IO Control call from the host. Passes a 
    arbitrary size of return data from a IO Control call back to the host.
\*--------------------------------------------------------------------------*/
void OtxDspCommandCompleteData(
  unsigned short hHandle, 
  unsigned short *pData,
  unsigned short dataLen
) 
{ 
  unsigned short i;
  unsigned short *pTempDataAddr = (unsigned short *)g_h2dDataAddrP;

  *g_h2dHandleAddrP  = hHandle;
  *g_h2dDataLenAddrP = dataLen;
  for (i = 0; i < dataLen; i++) {
    *(pTempDataAddr) = *pData++; 
    pTempDataAddr++;
  }
  *g_h2dCmdAddrP = OTXDSP_SDK_IO_NO_COMMAND; // Command completed 

  *C54X_HPIC_REG = C54X_HINTMASK; /* Assert HINT# towards the host */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
}

/*--------------------------------------------------------------------------*\
  @func
    Completes an IO Control call from the host. Passes a 
    return code from a IO Control call back to the host.
\*--------------------------------------------------------------------------*/
void OtxDspCommandCompleteRc(
  unsigned short hHandle,
  unsigned short nResult
)
{
  *g_h2dHandleAddrP  = hHandle;
  *g_h2dDataLenAddrP = sizeof(unsigned short);
  *g_h2dDataAddrP = nResult;
  *g_h2dCmdAddrP = OTXDSP_SDK_IO_NO_COMMAND; // Command completed

  *C54X_HPIC_REG = C54X_HINTMASK; /* Assert HINT# towards the host */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
}


/*--------------------------------------------------------------------------*\
  @func
    Completes an IO Control call from the host. No data is passed back
    to the host.
\*--------------------------------------------------------------------------*/
void OtxDspCommandComplete(
  unsigned short hHandle
)
{
  *g_h2dHandleAddrP  = hHandle;
  *g_h2dDataLenAddrP = 0;
  *g_h2dCmdAddrP = OTXDSP_SDK_IO_NO_COMMAND; // Command completed 

  *C54X_HPIC_REG = C54X_HINTMASK; /* Assert HINT# towards the host */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
}

/*--------------------------------------------------------------------------*\
  @func
    Easy access function for sending an event to the host with unsigned
    long as the data.
\*--------------------------------------------------------------------------*/
short OtxDspSetHostEventUlong(
  unsigned short nEvent,
  unsigned short hHandle,
  unsigned long data
)
{
  return OtxDspSetHostEventEx(nEvent, hHandle, OTXDSP_NO_TASKREF, (unsigned short *)&data, 2*sizeof(unsigned short), OTXDSP_FALSE);
}

/*--------------------------------------------------------------------------*\
  @func
    Easy access function for sending an event to the host with an unsigned
    short as the data.
\*--------------------------------------------------------------------------*/
short OtxDspSetHostEventUshort(
  unsigned short nEvent,
  unsigned short hHandle,
  unsigned short data
)
{
  return OtxDspSetHostEventEx(nEvent, hHandle, OTXDSP_NO_TASKREF, (unsigned short *)&data, sizeof(unsigned short), OTXDSP_FALSE);
}

/*--------------------------------------------------------------------------*\
  @func
    Easy access function for sending an event to the host with no data.
\*--------------------------------------------------------------------------*/
short OtxDspSetHostEventVoid(
  unsigned short nEvent,
  unsigned short hHandle
)
{
  return OtxDspSetHostEventEx(nEvent, hHandle, OTXDSP_NO_TASKREF, 0, 0, OTXDSP_FALSE);
}



/*--------------------------------------------------------------------------*\
  @func
    Generates a driver event for the host
\*--------------------------------------------------------------------------*/
short OtxDspSetHostEvent(
  unsigned short nEvent,         /* @parm Event to be send to the host */
  unsigned short hHandle,        /* @parm Handle to the user logical device sending the event */
  unsigned short nTaskRef,       /* @parm Associated Task Reference */
  unsigned short *pData,         /* @parm Pointer to the data to be sent to the host */
  unsigned short nDataLen        /* @parm Length of the data to be sent */
)
{
  return OtxDspSetHostEventEx(
    nEvent,         
    hHandle,        
    nTaskRef,       
    pData,         
    nDataLen,
    OTXDSP_FALSE   // nData is specified in words (not octets)
    );
}


/*--------------------------------------------------------------------------*\
  @func
\*--------------------------------------------------------------------------*/
inline void CalculateFreeBufferSpace(void)
{
  volatile unsigned short nCurrHostReadAddr = *((unsigned short *) OTXDSP_SDK_EVENT_BUF_PROCESSING_ADDR);

  if ((unsigned short)g_pCurrDspWriteAddr >= nCurrHostReadAddr) {
    g_hpiEventBufDiag.cwFreeBuffer = OTXDSP_SDK_EVENT_BUF_LEN - ((unsigned short)g_pCurrDspWriteAddr - nCurrHostReadAddr);
  } else {
    g_hpiEventBufDiag.cwFreeBuffer = nCurrHostReadAddr - (unsigned short)g_pCurrDspWriteAddr;
  }

  if (g_hpiEventBufDiag.cwFreeBuffer < g_hpiEventBufDiag.cwMinFreeBuffer) {
    g_hpiEventBufDiag.cwMinFreeBuffer = g_hpiEventBufDiag.cwFreeBuffer;
  }
}


/*--------------------------------------------------------------------------*\
  @func
\*--------------------------------------------------------------------------*/
void OtxDspCalculateFreeBufferSpace(void)
{
  CalculateFreeBufferSpace();
}



/*--------------------------------------------------------------------------*\
  @func
    Generates a driver event for the host

    Extended version. Has the ability to transfer a buffer of octets as
    well as a buffer of words.
\*--------------------------------------------------------------------------*/
short OtxDspSetHostEventEx(
  unsigned short nEvent,         /* @parm Event to be send to the host */
  unsigned short hHandle,        /* @parm Handle to the user logical device sending the event */
  unsigned short nTaskRef,       /* @parm Associated Task Reference */
  unsigned short *pData,         /* @parm Pointer to the data to be sent to the host */
  unsigned short nDataLen,       /* @parm Length of the data to be sent */
  short bOctetTransfer           /* @parm OTXDSP_TRUE if nDataLen and pData specifies octets. OTXDSP_FALSE if they specify words */
)
{
  unsigned short i;
  volatile unsigned short *pEventAddr;

  CalculateFreeBufferSpace();
  
  // If we have enough free buffer space then insert the command in the buffer
  if (g_hpiEventBufDiag.cwFreeBuffer >= (nDataLen 
    + 2 // +2 for the nEvent+NO_CMD
#ifdef OTXDSP_SDK_EVENT_BUF_INCLUDE_HANDLE
    + 1 // + 1 for handle
#endif
#ifdef OTXDSP_SDK_EVENT_BUF_INCLUDE_TASKREF
    + 1 // + 1 for taskref
#endif
#ifdef OTXDSP_SDK_EVENT_BUF_INCLUDE_TIMESTAMP
    + 4 // + 4 for timestamp
#endif
    )) 
  {   // +8 for the nEvent+handle+taskref+timestamp*4+NO_CMD
    // Don't write the event notify code yet. 
    // Just save the address where we want to write the code. 
    // We will write the event notify code when we have written the entire event packet 
    // to the buffer
    pEventAddr = g_pCurrDspWriteAddr; 

    g_pCurrDspWriteAddr++;
    if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
      g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
    }

#ifdef OTXDSP_SDK_EVENT_BUF_INCLUDE_HANDLE
    *g_pCurrDspWriteAddr = hHandle;

    g_pCurrDspWriteAddr++;
    if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
      g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
    }
#endif

#ifdef OTXDSP_SDK_EVENT_BUF_INCLUDE_TASKREF
    *g_pCurrDspWriteAddr = nTaskRef;

    g_pCurrDspWriteAddr++;
    if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
      g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
    }
#endif

#ifdef OTXDSP_SDK_EVENT_BUF_INCLUDE_TIMESTAMP
    // Time stamp
    {
      C54X_SET_INT_MASK_BIT;   // Globally disable interrupts by setting the INTM bit 

      *g_pCurrDspWriteAddr = g_timeStamp.nHighPart;
      
      g_pCurrDspWriteAddr++;
      if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
        g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
      }

      *g_pCurrDspWriteAddr = g_timeStamp.nMidPart;
      
      g_pCurrDspWriteAddr++;
      if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
        g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
      }

      *g_pCurrDspWriteAddr = g_timeStamp.nLowPart;
      
      g_pCurrDspWriteAddr++;
      if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
        g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
      }

      *g_pCurrDspWriteAddr = 0xFFFF - *C54X_TIM_REG;
      
      g_pCurrDspWriteAddr++;
      if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
        g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
      }

      C54X_CLR_INT_MASK_BIT;  // Globally enable interrupts by clearing the INTM bit
    }
#endif

    if (bOctetTransfer) {
      *(g_pCurrDspWriteAddr) = (nDataLen + 1) / 2; // data length
    } else {
      *(g_pCurrDspWriteAddr) = nDataLen; // data length
    }

    g_pCurrDspWriteAddr++;
    if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
      g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
    }

    if (bOctetTransfer) {
      // We will only transfer the LSB of the word buffer passed to this function
      for (i = 0; i < nDataLen/2; i++) {
        *(g_pCurrDspWriteAddr)  = *pData++ & 0x00FF; 
        *(g_pCurrDspWriteAddr) |= *pData++ << 8 ; 

        g_pCurrDspWriteAddr++;
        if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
          g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
        }
      }

      if (nDataLen & 0x0001) {
        // Transferring an odd number of octets. We are padding with an extra 0x00 octet
        nEvent |= OTXDSP_SDK_MASK_EVENT_DATA_PADDED;

        *(g_pCurrDspWriteAddr)  = *pData++ & 0x00FF; 

        g_pCurrDspWriteAddr++;
        if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
          g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
        }

      } 

    } else {
      for (i = 0; i < nDataLen; i++) {
        *(g_pCurrDspWriteAddr) = *pData++; 
        
        g_pCurrDspWriteAddr++;
        if (g_pCurrDspWriteAddr >= (unsigned short *)(OTXDSP_SDK_EVENT_BUF_ADDR + OTXDSP_SDK_EVENT_BUF_LEN)) {
          g_pCurrDspWriteAddr = (unsigned short *)OTXDSP_SDK_EVENT_BUF_ADDR;
        }
      }
    }

    *(g_pCurrDspWriteAddr) = OTXDSP_SDK_EVENT_NO_COMMAND; // We put this here so that host knows when to stop reading from the command buffer

    // Now it is safe to write the event notify code (now when the entire packet is written to the buffer)
    *(pEventAddr) = nEvent; 

  } else {
    // Buffer overrun. The host is too slow. Record this error condition
    g_hpiEventBufDiag.coLostEvents++;
    return 0; // Error
  }

#if CHIP_5410A || CHIP_54CST || CHIP_5416
  if ((*C54X_HPIC_REG) & C54X_HINTBIT) {
#if 0 // Do nothing if the bit is set
    unsigned k;
    for (k = 0; k < 10000; k++) {
      if ((*C54X_HPIC_REG) & C54X_HINTBIT) {
        // HINT is still set. Some delay and try to read it again
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
        C54X_NOP;     
      } else {
        // HINT was not set now. OK to proceed and set it.
        *C54X_HPIC_REG = C54X_HINTMASK; /* Assert HINT# towards the host */
        C54X_NOP;                  /* Let it sink in... */
        C54X_NOP;                  /* Let it sink in... */
        C54X_NOP;                  /* Let it sink in... */
        return 1; // success
      }
    }
    // HINT was never cleared. Error situation.
    g_hpiEventBufDiag.coLostEvents++;
    return 0; // Error
#endif
  } else {
    // HINT was not set. OK to proceed and set it.
    *C54X_HPIC_REG = C54X_HINTMASK; /* Assert HINT# towards the host */
    C54X_NOP;                  /* Let it sink in... */
    C54X_NOP;                  /* Let it sink in... */
    C54X_NOP;                  /* Let it sink in... */
  }
#else
  *C54X_HPIC_REG = C54X_HINTMASK; /* Assert HINT# towards the host */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
  C54X_NOP;                  /* Let it sink in... */
#endif

  return 1; // Success

}

